#!/usr/bin/env python3
#
#    Copyright (c) 2022 Project CHIP Authors
#    All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import argparse
import enum
import sys

import Base38
from bitarray import bitarray
from stdnum.verhoeff import calc_check_digit

# See section 5.1.4.1 Manual Pairing Code in the Matter specification v1.0
MANUAL_DISCRIMINATOR_LEN = 4
PINCODE_LEN = 27

MANUAL_CHUNK1_DISCRIMINATOR_MSBITS_LEN = 2
MANUAL_CHUNK1_DISCRIMINATOR_MSBITS_POS = 0
MANUAL_CHUNK1_VID_PID_PRESENT_BIT_POS = MANUAL_CHUNK1_DISCRIMINATOR_MSBITS_POS + MANUAL_CHUNK1_DISCRIMINATOR_MSBITS_LEN
MANUAL_CHUNK1_LEN = 1

MANUAL_CHUNK2_DISCRIMINATOR_LSBITS_LEN = 2
MANUAL_CHUNK2_PINCODE_LSBITS_LEN = 14
MANUAL_CHUNK2_PINCODE_LSBITS_POS = 0
MANUAL_CHUNK2_DISCRIMINATOR_LSBITS_POS = MANUAL_CHUNK2_PINCODE_LSBITS_POS + MANUAL_CHUNK2_PINCODE_LSBITS_LEN
MANUAL_CHUNK2_LEN = 5

MANUAL_CHUNK3_PINCODE_MSBITS_LEN = 13
MANUAL_CHUNK3_PINCODE_MSBITS_POS = 0
MANUAL_CHUNK3_LEN = 4

MANUAL_VID_LEN = 5
MANUAL_PID_LEN = 5

# See section 5.1.3. QR Code in the Matter specification v1.0
QRCODE_VERSION_LEN = 3
QRCODE_DISCRIMINATOR_LEN = 12
QRCODE_VID_LEN = 16
QRCODE_PID_LEN = 16
QRCODE_COMMISSIONING_FLOW_LEN = 2
QRCODE_DISCOVERY_CAP_BITMASK_LEN = 8
QRCODE_PADDING_LEN = 4
QRCODE_VERSION = 0
QRCODE_PADDING = 0

INVALID_PASSCODES = [00000000, 11111111, 22222222, 33333333, 44444444, 55555555,
                     66666666, 77777777, 88888888, 99999999, 12345678, 87654321]


class CommissioningFlow(enum.IntEnum):
    Standard = 0,
    UserIntent = 1,
    Custom = 2


class SetupPayload:
    def __init__(self, discriminator, pincode, rendezvous=4, flow=CommissioningFlow.Standard, vid=0, pid=0):
        self.long_discriminator = discriminator
        self.short_discriminator = discriminator >> 8
        self.pincode = pincode
        self.rendezvous = rendezvous
        self.flow = flow
        self.vid = vid
        self.pid = pid

    def manual_chunk1(self):
        discriminator_shift = (MANUAL_DISCRIMINATOR_LEN - MANUAL_CHUNK1_DISCRIMINATOR_MSBITS_LEN)
        discriminator_mask = (1 << MANUAL_CHUNK1_DISCRIMINATOR_MSBITS_LEN) - 1
        discriminator_chunk = (self.short_discriminator >> discriminator_shift) & discriminator_mask
        vid_pid_present_flag = 0 if self.flow == CommissioningFlow.Standard else 1
        return (discriminator_chunk << MANUAL_CHUNK1_DISCRIMINATOR_MSBITS_POS) | (vid_pid_present_flag << MANUAL_CHUNK1_VID_PID_PRESENT_BIT_POS)

    def manual_chunk2(self):
        discriminator_mask = (1 << MANUAL_CHUNK2_DISCRIMINATOR_LSBITS_LEN) - 1
        pincode_mask = (1 << MANUAL_CHUNK2_PINCODE_LSBITS_LEN) - 1
        discriminator_chunk = self.short_discriminator & discriminator_mask
        return ((self.pincode & pincode_mask) << MANUAL_CHUNK2_PINCODE_LSBITS_POS) | (discriminator_chunk << MANUAL_CHUNK2_DISCRIMINATOR_LSBITS_POS)

    def manual_chunk3(self):
        pincode_shift = PINCODE_LEN - MANUAL_CHUNK3_PINCODE_MSBITS_LEN
        pincode_mask = (1 << MANUAL_CHUNK3_PINCODE_MSBITS_LEN) - 1
        return ((self.pincode >> pincode_shift) & pincode_mask) << MANUAL_CHUNK3_PINCODE_MSBITS_POS

    def generate_manualcode(self):
        payload = str(self.manual_chunk1()).zfill(MANUAL_CHUNK1_LEN)
        payload += str(self.manual_chunk2()).zfill(MANUAL_CHUNK2_LEN)
        payload += str(self.manual_chunk3()).zfill(MANUAL_CHUNK3_LEN)

        if self.flow != CommissioningFlow.Standard:
            payload += str(self.vid).zfill(MANUAL_VID_LEN)
            payload += str(self.pid).zfill(MANUAL_PID_LEN)

        payload += calc_check_digit(payload)
        return payload

    def generate_qrcode(self):
        qrcode_bit_string = '{0:b}'.format(QRCODE_PADDING).zfill(QRCODE_PADDING_LEN)
        qrcode_bit_string += '{0:b}'.format(self.pincode).zfill(PINCODE_LEN)
        qrcode_bit_string += '{0:b}'.format(self.long_discriminator).zfill(QRCODE_DISCRIMINATOR_LEN)
        qrcode_bit_string += '{0:b}'.format(self.rendezvous).zfill(QRCODE_DISCOVERY_CAP_BITMASK_LEN)
        qrcode_bit_string += '{0:b}'.format(int(self.flow)).zfill(QRCODE_COMMISSIONING_FLOW_LEN)
        qrcode_bit_string += '{0:b}'.format(self.pid).zfill(QRCODE_PID_LEN)
        qrcode_bit_string += '{0:b}'.format(self.vid).zfill(QRCODE_VID_LEN)
        qrcode_bit_string += '{0:b}'.format(QRCODE_VERSION).zfill(QRCODE_VERSION_LEN)

        qrcode_bits = bitarray(qrcode_bit_string)
        bytes = list(qrcode_bits.tobytes())
        bytes.reverse()
        return 'MT:{}'.format(Base38.encode(bytes))


def validate_args(args):
    def check_int_range(value, min_value, max_value, name):
        if value and ((value < min_value) or (value > max_value)):
            print('{} is out of range, should be in range from {} to {}'.format(name, min_value, max_value))
            sys.exit(1)

    if args.passcode is not None:
        if ((args.passcode < 0x0000001 and args.passcode > 0x5F5E0FE) or (args.passcode in INVALID_PASSCODES)):
            print('Invalid passcode:' + str(args.passcode))
            sys.exit(1)

    check_int_range(args.discriminator, 0x0000, 0x0FFF, 'Discriminator')
    check_int_range(args.product_id, 0x0000, 0xFFFF, 'Product id')
    check_int_range(args.vendor_id, 0x0000, 0xFFFF, 'Vendor id')
    check_int_range(args.discovery_cap_bitmask, 0x0001, 0x0007, 'Discovery Capability Mask')


def main():
    def any_base_int(s): return int(s, 0)
    parser = argparse.ArgumentParser(description='Matter Manual and QRCode Setup Payload Generator Tool')
    parser.add_argument('-d', '--discriminator', type=any_base_int, required=True,
                              help='The discriminator for pairing, range: 0x00-0x0FFF')
    parser.add_argument('-p', '--passcode', type=any_base_int, required=True,
                              help='The setup passcode for pairing, range: 0x01-0x5F5E0FE')
    parser.add_argument('-vid', '--vendor-id', type=any_base_int, default=0, help='Vendor id')
    parser.add_argument('-pid', '--product-id', type=any_base_int, default=0, help='Product id')
    parser.add_argument('-cf', '--commissioning-flow', type=any_base_int, default=0,
                               help='Device commissioning flow, 0:Standard, 1:User-Intent, 2:Custom. \
                               Default is 0.', choices=[0, 1, 2])
    parser.add_argument('-dm', '--discovery-cap-bitmask', type=any_base_int, default=4,
                               help='Commissionable device discovery capability bitmask. \
                               0:SoftAP, 1:BLE, 2:OnNetwork. Default: OnNetwork')
    args = parser.parse_args()
    validate_args(args)

    payloads = SetupPayload(args.discriminator, args.passcode, args.discovery_cap_bitmask,
                            CommissioningFlow(args.commissioning_flow), args.vendor_id, args.product_id)
    manualcode = payloads.generate_manualcode()
    qrcode = payloads.generate_qrcode()

    print("Manualcode : {}".format(manualcode))
    print("QRCode     : {}".format(qrcode))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

#
#    Copyright (c) 2022 Project CHIP Authors
#    All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import argparse
import base64
import hashlib
import struct

from ecdsa.curves import NIST256p

# Forbidden passcodes as listed in the "5.1.7.1. Invalid Passcodes" section of the Matter spec
INVALID_PASSCODES = [00000000,
                     11111111,
                     22222222,
                     33333333,
                     44444444,
                     55555555,
                     66666666,
                     77777777,
                     88888888,
                     99999999,
                     12345678,
                     87654321, ]

# Length of `w0s` and `w1s` elements
WS_LENGTH = NIST256p.baselen + 8


def generate_verifier(passcode: int, salt: bytes, iterations: int) -> bytes:
    ws = hashlib.pbkdf2_hmac('sha256', struct.pack('<I', passcode), salt, iterations, WS_LENGTH * 2)
    w0 = int.from_bytes(ws[:WS_LENGTH], byteorder='big') % NIST256p.order
    w1 = int.from_bytes(ws[WS_LENGTH:], byteorder='big') % NIST256p.order
    L = NIST256p.generator * w1

    return w0.to_bytes(NIST256p.baselen, byteorder='big') + L.to_bytes('uncompressed')


def main():
    def passcode_arg(arg: str) -> int:
        passcode = int(arg)

        if not 0 <= passcode <= 99999999:
            raise argparse.ArgumentTypeError('passcode out of range')

        if passcode in INVALID_PASSCODES:
            raise argparse.ArgumentTypeError('invalid passcode')

        return passcode

    def salt_arg(arg: str) -> bytes:
        salt = base64.b64decode(arg)

        if not 16 <= len(salt) <= 32:
            raise argparse.ArgumentTypeError('invalid salt length')

        return salt

    def iterations_arg(arg: str) -> int:
        iterations = int(arg)

        if not 1000 <= iterations <= 100000:
            raise argparse.ArgumentTypeError('iteration count out of range')

        return iterations

    parser = argparse.ArgumentParser(description='SPAKE2+ Python Tool', fromfile_prefix_chars='@')
    commands = parser.add_subparsers(dest='command', metavar='subcommand'.ljust(16), required=True)

    gen_verifier = commands.add_parser('gen-verifier', help='Generate SPAKE2+ Verifier')
    gen_verifier.add_argument('-p', '--passcode', type=passcode_arg,
                              required=True, help='8-digit passcode')
    gen_verifier.add_argument('-s', '--salt', type=salt_arg,
                              required=True, help='Salt of length 16 to 32 octets encoded in Base64')
    gen_verifier.add_argument('-i', '--iteration-count', type=iterations_arg,
                              metavar='count', required=True, help='Iteration count between 1000 and 100000')

    args = parser.parse_args()

    if args.command == 'gen-verifier':
        verifier = generate_verifier(args.passcode, args.salt, args.iteration_count)
        print(base64.b64encode(verifier).decode('ascii'))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# Copyright 2024 Espressif Systems (Shanghai) PTE LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
contains utility functions for generating csr, build certificates, certs/keys conversion etc.
"""
import datetime
from typing import Optional
from cryptography import x509
from datetime import datetime, timedelta
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.x509 import load_pem_x509_certificate, load_der_x509_certificate

# CHIP OID for vendor id
VENDOR_ID_OID = x509.ObjectIdentifier('1.3.6.1.4.1.37244.2.1')

# CHIP OID for product id
PRODUCT_ID_OID = x509.ObjectIdentifier('1.3.6.1.4.1.37244.2.2')

VALID_DAYS = 365 * 100

def save_to_file(file_name, file_data):
    """
    Save filedata to file
    Args:
    file_name (str): file name for the file.
    file_data (bytes): file data to store in file.
    """
    with open(file_name, 'wb') as file:
        file.write(file_data)

def chip_gen_ec_key():
    """
    Generate EC private key.
    Returns:
        ec.EllipticCurvePrivateKey: Generated private key object.
    """
    return ec.generate_private_key(ec.SECP256R1())

def load_cert_from_file(cert_file):
    """
    Load an X.509 certificate from file in PEM or DER format.
    Args:
        cert_file (str): Path to the certificate file.
    Returns:
        x509.Certificate: Parsed certificate object in PEM or DER format.
    Raises:
        ValueError: If the certificate format is invalid or unsupported.
    """
    with open(cert_file, 'rb') as file:
        cert_data = file.read()

    # Attempt to load as PEM first, then fallback to DER
    for loader in (load_pem_x509_certificate, load_der_x509_certificate):
        try:
            return loader(cert_data)
        except ValueError:
            continue

    raise ValueError("Invalid certificate format. Supported formats: PEM, DER.")

def load_key_from_file(key_file):
    """
    Load a private key from file in PEM or DER format.
    Args:
        key_file (str): Path to the private key file.
    Returns:
        ec.EllipticCurvePrivateKey: Parsed private key object in PEM or DER format.
    Raises:
        ValueError: If the key format is invalid or unsupported.
    """
    with open(key_file, 'rb') as file:
        key_data = file.read()

    # Attempt to load as PEM first, then fallback to DER
    for loader in (serialization.load_pem_private_key, serialization.load_der_private_key):
        try:
            return loader(key_data, password=None)
        except ValueError:
            continue

    raise ValueError("Invalid key format or unsupported key type. Supported formats: PEM, DER.")

def convert_x509_cert_from_pem_to_der(pem_file, out_der_file):
    """
    Convert an X509 certificate from PEM to DER format.
    Args:
        pem_file (str): Path to the PEM file.
        out_der_file (str): Output path for the DER file.
    """
    cert = load_cert_from_file(pem_file)
    der_cert = cert.public_bytes(serialization.Encoding.DER)
    save_to_file(out_der_file, der_cert)

def convert_private_key_from_pem_to_der(pem_file, out_der_file):
    """
    Convert a private key from PEM to DER format.
    Args:
        pem_file (str): Path to the PEM file.
        out_der_file (str): Output path for the DER file.
    """
    key = load_key_from_file(pem_file)
    der_key = key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )
    save_to_file(out_der_file, der_key)

def store_keypair_as_raw(pem_file, out_privkey_bin, out_pubkey_bin):
    """
    Generate a binary format key pair from a PEM file.
    Args:
        pem_file (str): Path to the PEM file containing the private key.
        out_privkey_bin (str): Output path for the private key in binary format.
        out_pubkey_bin (str): Output path for the public key in binary format.
    """
    key = load_key_from_file(pem_file)
    priv_val = key.private_numbers().private_value
    pub_numbers = key.public_key().public_numbers()

    priv_key_data = priv_val.to_bytes(32, byteorder='big')
    save_to_file(out_privkey_bin, priv_key_data)

    with open(out_pubkey_bin, 'wb') as pub_file:
        pub_file.write(b'\x04')
        pub_file.write(pub_numbers.x.to_bytes(32, byteorder='big'))
        pub_file.write(pub_numbers.y.to_bytes(32, byteorder='big'))

def extract_matter_rdn(cert, oid):
    """
    Extract the value of a custom OID from a certificate.
    Args:
        cert (x509.Certificate): Certificate from which to extract the OID value.
        oid (x509.ObjectIdentifier): OID to extract.
    Returns:
        str: Extracted OID value or None if not found.
    """
    try:
        return cert.subject.get_attributes_for_oid(oid)[0].value
    except IndexError:
        return None

def extract_pid(cert):
    """
    Extract the Product ID (PID) from a certificate's subject.
    Args:
        cert (x509.Certificate): Certificate from which to extract the PID.
    Returns:
        str: Extracted PID or None if not found.
    """
    return extract_matter_rdn(cert, PRODUCT_ID_OID)

def extract_vid(cert):
    """
    Extract the Vendor ID (VID) from a certificate's subject.
    Args:
        cert (x509.Certificate): Certificate from which to extract the VID.
    Returns:
        str: Extracted VID or None if not found.
    """
    return extract_matter_rdn(cert, VENDOR_ID_OID)

def extract_common_name(cert: x509.Certificate) -> Optional[str]:
    """
    Extract the Common Name (CN) from a certificate's subject.
    Args:
        cert (x509.Certificate): Certificate from which to extract the CN.
    Returns:
        str: Extracted CN or None if not found.
    """
    return extract_matter_rdn(cert, x509.NameOID.COMMON_NAME)

def generate_cert_subject(vendor_id, product_id, common_name):
    x509_attrs = []
    x509_attrs.append(x509.NameAttribute(NameOID.COMMON_NAME, common_name))
    x509_attrs.append(x509.NameAttribute(VENDOR_ID_OID, vendor_id.zfill(4)))
    x509_attrs.append(x509.NameAttribute(PRODUCT_ID_OID, product_id.zfill(4)))
    return x509.Name(x509_attrs)

def build_certificate(vendor_id:str, product_id:str, ca_cert_file: str, ca_privkey_file: str, out_cert_file: str, out_key_file: str, is_pai: bool, common_name: str, valid_from=None, lifetime=None):
    """
    Build a certificate for a Matter device.

    Args:
        vendor_id (str): vendor id
        product_id (str): product id
        ca_cert_file (str): Path to the CA certificate file used for signing.
        ca_privkey_file (str): Path to the CA private key file used for signing.
        out_cert_file (str): Path to save the generated certificate.
        out_key_file (str): Path to save the private key for the certificate.
        is_pai (bool): Indicates if the certificate is a Product Attestation Intermediate (PAI) certificate.
        common_name (str): Common Name for the certificate subject.
        valid_from (str, optional): Start date for the certificate validity period in ISO 8601 format ("YYYY-MM-DDTHH:MM:SS").
            Defaults to None, in which case the current date is used.
        lifetime (int, optional): Lifetime of the certificate in days. Defaults to None, in which case a default period (e.g., 100 years) is used.

    Returns:
        None: Saves the generated certificate and key to the specified files.
    """
    if not ca_cert_file or not ca_privkey_file:
        raise ValueError("CA key and certificate cannot be None.")

    try:
        # Load CA certificate and key
        ca_cert = load_cert_from_file(ca_cert_file)
        ca_key = load_key_from_file(ca_privkey_file)

        # Generate a new private key for the device certificate
        private_key = chip_gen_ec_key()
        public_key = private_key.public_key()

        # Build subject attributes
        cert_subject = generate_cert_subject(vendor_id, product_id, common_name)

        # Define validity period
        nvb_time = datetime.utcnow()
        nva_time = nvb_time + timedelta(days=VALID_DAYS)  # Default to 100 years
        if valid_from:
            nvb_time = datetime.strptime(str(valid_from), "%Y-%m-%dT%H:%M:%S")
        if lifetime:
            nva_time = nvb_time + timedelta(days=lifetime)

        # Build certificate
        cert = x509.CertificateBuilder()
        cert = cert.subject_name(cert_subject)
        cert = cert.issuer_name(ca_cert.subject)
        cert = cert.public_key(public_key)
        cert = cert.serial_number(x509.random_serial_number())
        cert = cert.not_valid_before(nvb_time)
        cert = cert.not_valid_after(nva_time)

        # Extensions for PAI certificate
        if is_pai:
            cert = cert.add_extension(x509.BasicConstraints(ca=True, path_length=0), critical=True)
            cert = cert.add_extension(x509.KeyUsage(digital_signature=True, content_commitment=False,
                                                    key_encipherment=False, data_encipherment=False,
                                                    key_agreement=False, key_cert_sign=True, crl_sign=True,
                                                    encipher_only=False, decipher_only=False), critical=True)
        # Extensions for DAC certificate
        else:
            cert = cert.add_extension(x509.BasicConstraints(ca=False, path_length=None), critical=True)
            cert = cert.add_extension(x509.KeyUsage(digital_signature=True, content_commitment=False,
                                                    key_encipherment=False, data_encipherment=False,
                                                    key_agreement=False, key_cert_sign=False, crl_sign=False,
                                                    encipher_only=False, decipher_only=False), critical=True)

        cert = cert.add_extension(x509.SubjectKeyIdentifier.from_public_key(public_key), critical=False)
        cert = cert.add_extension(x509.AuthorityKeyIdentifier.from_issuer_public_key(ca_key.public_key()), critical=False)

        # Sign the certificate
        signed_cert = cert.sign(private_key=ca_key, algorithm=hashes.SHA256(), backend=default_backend())

        # Save the certificate to a file
        certificate_pem = signed_cert.public_bytes(serialization.Encoding.PEM)
        save_to_file(out_cert_file, certificate_pem)

        # Save the private key to a file
        private_key_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        )
        save_to_file(out_key_file, private_key_pem)

    except Exception as e:
        raise RuntimeError(f"Failed to build certificate: {e}")



#!/usr/bin/env python3

# Copyright 2022 Espressif Systems (Shanghai) PTE LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Script to generate Matter factory NVS partition image, Onboarding codes, and QR codes.
"""

import os
import sys
import csv
import uuid
import shutil
import base64
import random
import logging
import binascii
import argparse
import pyqrcode

from chip_nvs import *
from utils import *
from datetime import datetime
from types import SimpleNamespace
from cert_utils import *

from esp_secure_cert import configure_ds
from esp_secure_cert.tlv_format import *

# In order to made the esp-matter-mfg-tool standalone we copied few dependencies from esp-idf
# and connectedhomeip to deps/ directory.
# TODO: Remove the dependencies from deps/ once available on pypi
from deps.spake2p import generate_verifier
from deps.mfg_gen import generate
from deps.generate_setup_payload import SetupPayload, CommissioningFlow

PAI = {
    'cert_pem': None,
    'cert_der': None,
    'key_pem': None,
    'key_der': None,
}

OUT_DIR = {
    'top': None,
    'chip': None,
}

OUT_FILE = {
    'config_csv': None,
    'mcsv': None,
    'pin_csv': None,
    'pin_disc_csv': None,
    'cn_dac_csv': None
}

# Supported log levels, mapping string values required for argument
# parsing into logging constants
__LOG_LEVELS__ = {
    'info': logging.INFO,
    'error': logging.ERROR,
}

UUIDs = list()

def generate_passcodes(args):
    iter_count_max = 10000
    salt_len_max = 32
    with open(OUT_FILE['pin_csv'], 'w', newline='') as f:
        writer = csv.writer(f)
        if args.enable_dynamic_passcode:
            writer.writerow(["Index", "Iteration Count", "Salt"])
        else:
            writer.writerow(["Index", "PIN Code", "Iteration Count", "Salt", "Verifier"])
        for i in range(0, args.count):
            salt = os.urandom(salt_len_max)
            if args.enable_dynamic_passcode:
                writer.writerow([i, iter_count_max, base64.b64encode(salt).decode('utf-8')])
            else:
                if args.passcode:
                    passcode = args.passcode
                else:
                    passcode = random.randint(1, 99999998)
                    if passcode in INVALID_PASSCODES:
                        passcode -= 1
                verifier = generate_verifier(passcode, salt, iter_count_max)
                writer.writerow([i, passcode, iter_count_max, base64.b64encode(salt).decode('utf-8'), base64.b64encode(verifier).decode('utf-8')])


def generate_discriminators(args):
    discriminators = list()

    # If discriminator is provided, use it
    if args.discriminator:
        discriminators.append(args.discriminator)
    else:
        for i in range(args.count):
            discriminators.append(random.randint(0x0000, 0x0FFF))

    return discriminators


# Append discriminator to each line of the passcode file
def append_discriminator(discriminator):
    with open(OUT_FILE['pin_csv'], 'r') as fd:
        lines = fd.readlines()

    lines[0] = ','.join([lines[0].strip(), 'Discriminator'])
    for i in range(1, len(lines)):
        lines[i] = ','.join([lines[i].strip(), str(discriminator[i - 1])])

    with open(OUT_FILE['pin_disc_csv'], 'w') as fd:
        fd.write('\n'.join(lines) + '\n')

    os.remove(OUT_FILE['pin_csv'])


# Generates the csv file containing chip specific keys and keys provided by user in csv file
def generate_config_csv(args):
    logging.info("Generating Config CSV...")
    csv_data = chip_nvs_get_config_csv()

    with open(OUT_FILE['config_csv'], 'w') as f:
        f.write(csv_data)


def write_chip_mcsv_header(args):
    logging.info('Writing chip manifest CSV header...')
    mcsv_header = chip_get_keys_as_csv() + '\n'
    with open(OUT_FILE['mcsv'], 'w') as f:
        f.write(mcsv_header)


def append_chip_mcsv_row(row_data):
    logging.info('Appending chip master CSV row...')
    with open(OUT_FILE['mcsv'], 'a') as f:
        f.write(row_data + '\n')

def generate_pai(args, ca_key, ca_cert, out_key, out_cert):
    vendor_id = hex(args.vendor_id)[2:].upper()
    product_id = hex(args.product_id)[2:].upper()
    common_name = "{} PAI {}".format(args.cn_prefix, "00") if args.cn_prefix else "MATTER TEST PAI " + str(random.randint(10, 99))

    build_certificate(
        vendor_id=vendor_id,
        product_id=product_id,
        ca_cert_file=ca_cert,
        ca_privkey_file=ca_key,
        out_cert_file=out_cert,
        out_key_file=out_key,
        is_pai=True,
        common_name=common_name,
        valid_from=args.valid_from,
        lifetime=args.lifetime)

    logging.info('Generated PAI certificate: {}'.format(out_cert))
    logging.info('Generated PAI private key: {}'.format(out_key))


def generate_dac(iteration, args, ca_key, ca_cert):
    out_key_pem = os.sep.join([OUT_DIR['top'], UUIDs[iteration], 'internal', 'DAC_key.pem'])
    out_private_key_der = out_key_pem.replace('key.pem', 'key.der')
    out_cert_pem = out_key_pem.replace('key.pem', 'cert.pem')
    out_cert_der = out_key_pem.replace('key.pem', 'cert.der')
    out_private_key_bin = out_key_pem.replace('key.pem', 'private_key.bin')
    out_public_key_bin = out_key_pem.replace('key.pem', 'public_key.bin')

    vendor_id = hex(args.vendor_id)[2:].upper()
    product_id = hex(args.product_id)[2:].upper()
    common_name = UUIDs[iteration]

    build_certificate(
        vendor_id=vendor_id,
        product_id=product_id,
        ca_cert_file=ca_cert,
        ca_privkey_file=ca_key,
        out_cert_file=out_cert_pem,
        out_key_file=out_key_pem,
        is_pai=False,
        common_name=common_name,
        valid_from=args.valid_from,
        lifetime=args.lifetime)

    logging.info('Generated DAC certificate: {}'.format(out_cert_pem))
    logging.info('Generated DAC private key: {}'.format(out_key_pem))

    convert_x509_cert_from_pem_to_der(out_cert_pem, out_cert_der)
    logging.info('Generated DAC certificate in DER format: {}'.format(out_cert_der))

    store_keypair_as_raw(out_key_pem, out_private_key_bin, out_public_key_bin)
    logging.info('Generated DAC private key in binary format: {}'.format(out_private_key_bin))
    logging.info('Generated DAC public key in binary format: {}'.format(out_public_key_bin))
    convert_private_key_from_pem_to_der(out_key_pem, out_private_key_der)
    return out_cert_der, out_private_key_bin, out_public_key_bin, out_private_key_der


def use_dac_from_args(args):
    logging.info('Using DAC from command line arguments...')
    logging.info('DAC Certificate: {}'.format(args.dac_cert))
    logging.info('DAC Private Key: {}'.format(args.dac_key))

    # There should be only one UUID in the UUIDs list if DAC is specified
    out_cert_der = os.sep.join([OUT_DIR['top'], UUIDs[0], 'internal', 'DAC_cert.der'])
    out_private_key_bin = out_cert_der.replace('cert.der', 'private_key.bin')
    out_public_key_bin = out_cert_der.replace('cert.der', 'public_key.bin')
    out_private_key_der = out_cert_der.replace('cert.der', 'key.der')

    convert_x509_cert_from_pem_to_der(args.dac_cert, out_cert_der)
    logging.info('Generated DAC certificate in DER format: {}'.format(out_cert_der))

    store_keypair_as_raw(args.dac_key, out_private_key_bin, out_public_key_bin)
    logging.info('Generated DAC private key in binary format: {}'.format(out_private_key_bin))
    logging.info('Generated DAC public key in binary format: {}'.format(out_public_key_bin))
    convert_private_key_from_pem_to_der(args.dac_key, out_private_key_der)

    return out_cert_der, out_private_key_bin, out_public_key_bin, out_private_key_der


def is_valid_uuid(uuid_str: str) -> bool:
    try:
        uuid.UUID(uuid_str)
        return True
    except ValueError:
        return False


def setup_out_dirs(vid, pid, count, outdir, arg_dac_cert):
    OUT_DIR['top'] = os.sep.join([outdir, vid_pid_str(vid, pid)])
    OUT_DIR['stage'] = os.sep.join([outdir, vid_pid_str(vid, pid), 'staging'])

    os.makedirs(OUT_DIR['top'], exist_ok=True)
    os.makedirs(OUT_DIR['stage'], exist_ok=True)

    OUT_FILE['config_csv'] = os.sep.join([OUT_DIR['stage'], 'config.csv'])
    OUT_FILE['mcsv'] = os.sep.join([OUT_DIR['stage'], 'master.csv'])
    OUT_FILE['pin_csv'] = os.sep.join([OUT_DIR['stage'], 'pin.csv'])
    OUT_FILE['pin_disc_csv'] = os.sep.join([OUT_DIR['stage'], 'pin_disc.csv'])
    OUT_FILE['cn_dac_csv'] = os.sep.join([OUT_DIR['top'], 'cn_dacs-{}.csv'.format(datetime.now().strftime("%Y-%m-%d-%H-%M-%S-%f"))])

    # If user provided the DAC, then count is 1 and this gets called only after
    # arguments have been validated. So we can safely assume that count is 1, and return
    if arg_dac_cert:
        subject_cn = extract_common_name(load_cert_from_file(arg_dac_cert))
        if subject_cn and is_valid_uuid(subject_cn):
            UUIDs.append(subject_cn)
            os.makedirs(os.sep.join([OUT_DIR['top'], subject_cn, 'internal']), exist_ok=True)
        return

    # Create directories to store the generated files
    for i in range(count):
        uuid_str = str(uuid.uuid4())
        UUIDs.append(uuid_str)
        os.makedirs(os.sep.join([OUT_DIR['top'], uuid_str, 'internal']), exist_ok=True)


def generate_passcodes_and_discriminators(args):
    # Generate passcodes using spake2p tool
    generate_passcodes(args)
    # Randomly generate discriminators
    discriminators = generate_discriminators(args)
    # Append discriminators to passcodes file
    append_discriminator(discriminators)


def write_cn_dac_csv_header():
    with open(OUT_FILE['cn_dac_csv'], 'a') as csv_file:
        csv_file.write("CN,certs\n")
    return

def write_csv_files(args):
    generate_config_csv(args)
    write_chip_mcsv_header(args)
    write_cn_dac_csv_header()


def setup_root_certs(args):
    # If PAA is passed as input, then generate PAI certificate
    if args.paa:
        # output file names
        PAI['cert_pem'] = os.sep.join([OUT_DIR['stage'], 'pai_cert.pem'])
        PAI['cert_der'] = os.sep.join([OUT_DIR['stage'], 'pai_cert.der'])
        PAI['key_pem'] = os.sep.join([OUT_DIR['stage'], 'pai_key.pem'])

        generate_pai(args, args.key, args.cert, PAI['key_pem'], PAI['cert_pem'])
        convert_x509_cert_from_pem_to_der(PAI['cert_pem'], PAI['cert_der'])
        logging.info('Generated PAI certificate in DER format: {}'.format(PAI['cert_der']))

    # If PAI is passed as input, generate DACs
    elif args.pai:
        PAI['cert_pem'] = args.cert
        PAI['key_pem'] = args.key
        PAI['cert_der'] = os.sep.join([OUT_DIR['stage'], 'pai_cert.der'])

        convert_x509_cert_from_pem_to_der(PAI['cert_pem'], PAI['cert_der'])
        logging.info('Generated PAI certificate in DER format: {}'.format(PAI['cert_der']))


def overwrite_values_in_mcsv(args, index):
    with open(args.mcsv, 'r') as mcsvf:
        mcsv_dict = list(csv.DictReader(mcsvf))[index]
        with open(args.csv, 'r') as csvf:
            csv_reader = csv.reader(csvf, delimiter=',')
            current_namespace = 'chip-factory'
            for csv_data in csv_reader:
                if 'namespace' in csv_data:
                    current_namespace = csv_data[0]
                else:
                    chip_nvs_map_update(current_namespace, csv_data[0], csv_data[1], csv_data[2], mcsv_dict[csv_data[0]])


def append_cn_dac_to_csv(common_name, cert_path):
    with open(OUT_FILE['cn_dac_csv'], 'a') as csv_file:
        device_cert_contents = load_cert_from_file(cert_path).public_bytes(serialization.Encoding.PEM).decode('utf-8')
        csv_file.write('{},"{}"\n'.format(common_name, device_cert_contents))

# This function generates the DACs, picks the commissionable data from the already present csv file,
# and generates the onboarding payloads, and writes everything to the master csv
def write_per_device_unique_data(args):
    with open(OUT_FILE['pin_disc_csv'], 'r') as csvf:
        pin_disc_dict = csv.DictReader(csvf)

        for row in pin_disc_dict:
            chip_factory_update('discriminator', row['Discriminator'])
            chip_factory_update('iteration-count', row['Iteration Count'])
            chip_factory_update('salt', row['Salt'])
            if not args.enable_dynamic_passcode:
                chip_factory_update('verifier', row['Verifier'])
            if args.paa or args.pai:
                if args.dac_key is not None and args.dac_cert is not None:
                    dacs = use_dac_from_args(args)
                else:
                    dacs = generate_dac(int(row['Index']), args, PAI['key_pem'], PAI['cert_pem'])

                if not args.dac_in_secure_cert:
                    chip_factory_update('dac-cert', os.path.abspath(dacs[0]))
                    chip_factory_update('dac-key', os.path.abspath(dacs[1]))
                    chip_factory_update('dac-pub-key', os.path.abspath(dacs[2]))
                    chip_factory_update('pai-cert', os.path.abspath(PAI['cert_der']))
                else:
                # esp secure cert partition
                    secure_cert_partition_file_path = os.sep.join([OUT_DIR['top'], UUIDs[int(row['Index'])], UUIDs[int(row['Index'])] + '_esp_secure_cert.bin'])
                    if args.ds_peripheral:
                        if args.target != "esp32h2":
                            logging.error("DS peripheral is only supported for esp32h2 target")
                            exit(1)

                        if args.efuse_key_id == -1:
                            logging.error("--efuse-key-id <value> is required when -ds or --ds-peripheral option is used")
                            exit(1)

                        if args.port and args.count == 1:
                            esp_secure_cert_data_dir = 'esp_secure_cert_data'
                            if (os.path.exists(esp_secure_cert_data_dir) is False):
                                os.makedirs(esp_secure_cert_data_dir)
                            ecdsa_key_file = os.path.join(esp_secure_cert_data_dir, 'ecdsa_key.bin')

                            ecdsa_key_size = '256'
                            configure_ds.configure_efuse_for_ecdsa(args.target, args.port, ecdsa_key_file, None, esp_secure_cert_data_dir, ecdsa_key_size, os.path.abspath(dacs[3]), args.priv_key_pass, args.efuse_key_id)
                        else:
                            logging.error('Port not specified or number of partitions count is greater than 1')
                            exit(1)

                        priv_key = tlv_priv_key_t(key_type = tlv_priv_key_type_t.ESP_SECURE_CERT_ECDSA_PERIPHERAL_KEY,
                                                  key_path = os.path.abspath(dacs[3]), key_pass = None)
                        priv_key.priv_key_len = 256
                        priv_key.efuse_key_id = args.efuse_key_id
                        generate_partition_ds(priv_key = priv_key, device_cert = os.path.abspath(dacs[0]),
                                              ca_cert = os.path.abspath(PAI['cert_der']), idf_target = args.target,
                                              op_file = secure_cert_partition_file_path)
                    else:
                        priv_key = tlv_priv_key_t(key_type = tlv_priv_key_type_t.ESP_SECURE_CERT_DEFAULT_FORMAT_KEY,
                                                  key_path = os.path.abspath(dacs[3]), key_pass = None)
                        generate_partition_no_ds(priv_key = priv_key, device_cert = os.path.abspath(dacs[0]),
                                                 ca_cert = os.path.abspath(PAI['cert_der']), idf_target = args.target,
                                                 op_file = secure_cert_partition_file_path)

                # appends the subject's common name and DAC certificate encoded as PEM to the csv file
                file_name = os.sep.join([OUT_DIR['top'], UUIDs[int(row['Index'])], "internal", "DAC_cert.pem"])
                if args.dac_key is not None and args.dac_cert is not None:
                    file_name = args.dac_cert

                subject_cn = extract_common_name(load_cert_from_file(file_name))
                # If common name is not present, lets skip adding that entry to the csv file
                if subject_cn:
                    append_cn_dac_to_csv(subject_cn, file_name)
                else:
                    logging.warning("Skipping entry for device with index {} as common name is not present in the DAC certificate".format(row['Index']))

            # If serial number is not passed, then generate one
            if (args.serial_num is None):
                chip_factory_update('serial-num', binascii.b2a_hex(os.urandom(SERIAL_NUMBER_LEN)).decode('utf-8'))

            if (args.enable_rotating_device_id is True) and (args.rd_id_uid is None):
                chip_factory_update('rd-id-uid', binascii.b2a_hex(os.urandom(int(ROTATING_DEVICE_ID_UNIQUE_ID_LEN_BITS / 8))).decode('utf-8'))

            if (args.csv is not None and args.mcsv is not None):
                overwrite_values_in_mcsv(args, int(row['Index']))

            mcsv_row_data = chip_get_values_as_csv()
            append_chip_mcsv_row(mcsv_row_data)

            # Generate onboarding data
            if not args.enable_dynamic_passcode:
                generate_onboarding_data(args, int(row['Index']), int(chip_factory_get_val('discriminator')), int(row['PIN Code']))
        if args.paa or args.pai:
            logging.info("Generated CSV of Common Name and DAC: {}".format(OUT_FILE['cn_dac_csv']))


def organize_output_files(suffix, args):
    for i in range(len(UUIDs)):
        dest_path = os.sep.join([OUT_DIR['top'], UUIDs[i]])
        internal_path = os.sep.join([dest_path, 'internal'])

        if args.generate_bin:
            replace = os.sep.join([OUT_DIR['top'], 'bin', '{}-{}.bin'.format(suffix, str(i + 1))])
            replace_with = os.sep.join([dest_path, '{}-partition.bin'.format(UUIDs[i])])
            os.rename(replace, replace_with)

            if args.encrypt:
                 replace = os.sep.join([OUT_DIR['top'], 'keys', 'keys-{}-{}.bin'.format(suffix, str(i + 1))])
                 replace_with = os.sep.join([dest_path, '{}-keys-partition.bin'.format(UUIDs[i])])
                 os.rename(replace, replace_with)

        replace = os.sep.join([OUT_DIR['top'], 'csv', '{}-{}.csv'.format(suffix, str(i + 1))])
        replace_with = os.sep.join([internal_path, 'partition.csv'])
        os.rename(replace, replace_with)

        # Also copy the PAI certificate to the output directory
        if args.paa or args.pai:
            shutil.copy2(PAI['cert_der'], os.sep.join([internal_path, 'PAI_cert.der']))

        logging.info('Generated output files at: {}'.format(os.sep.join([OUT_DIR['top'], UUIDs[i]])))

    shutil.rmtree(os.sep.join([OUT_DIR['top'], 'bin']), ignore_errors=True)
    shutil.rmtree(os.sep.join([OUT_DIR['top'], 'csv']), ignore_errors=True)

    if args.encrypt:
        shutil.rmtree(os.sep.join([OUT_DIR['top'], 'keys']), ignore_errors=True)

def generate_summary(args):
    master_csv = os.sep.join([OUT_DIR['stage'], 'master.csv'])
    summary_csv = os.sep.join([OUT_DIR['top'], 'summary-{}.csv'.format(datetime.now().strftime("%Y-%m-%d-%H-%M-%S"))])

    summary_csv_data = ''
    with open(master_csv, 'r') as mcsvf:
        summary_lines = mcsvf.read().splitlines()
        summary_csv_data += summary_lines[0]
        if not args.enable_dynamic_passcode:
            summary_csv_data += ',pincode,qrcode,manualcode\n'
        else:
            summary_csv_data += '\n'
        with open(OUT_FILE['pin_disc_csv'], 'r') as pdcsvf:
            pin_disc_dict = csv.DictReader(pdcsvf)
            for row in pin_disc_dict:
                if not args.enable_dynamic_passcode:
                    pincode = row['PIN Code']
                    discriminator = row['Discriminator']
                    payloads = SetupPayload(int(discriminator), int(pincode), args.discovery_mode, CommissioningFlow(args.commissioning_flow),
                                            args.vendor_id, args.product_id)
                    qrcode = payloads.generate_qrcode()
                    manualcode = payloads.generate_manualcode()
                    # ToDo: remove this if qrcode tool can handle the standard manual code format
                    if args.commissioning_flow == CommissioningFlow.Standard:
                        manualcode = manualcode[:4] + '-' + manualcode[4:7] + '-' + manualcode[7:]
                    else:
                        manualcode = '"' + manualcode[:4] + '-' + manualcode[4:7] + '-' + manualcode[7:11] + '\n' + manualcode[11:15] + '-' + manualcode[15:18] + '-' + manualcode[18:20] + '-' + manualcode[20:21] + '"'
                    summary_csv_data += summary_lines[1 + int(row['Index'])] + ',' + pincode + ',' + qrcode + ',' + manualcode + '\n'
                else:
                    summary_csv_data += summary_lines[1 + int(row['Index'])] + '\n'

    with open(summary_csv, 'w') as scsvf:
        scsvf.write(summary_csv_data)

def generate_partitions(suffix, size, encrypt, generate_partition_bin):
    partition_args = SimpleNamespace(fileid = None,
                                     version = 2,
                                     inputkey = None,
                                     outdir = OUT_DIR['top'],
                                     conf = OUT_FILE['config_csv'],
                                     values = OUT_FILE['mcsv'],
                                     size = hex(size),
                                     prefix = suffix,
                                     generate_bin = generate_partition_bin)
    if encrypt:
        partition_args.keygen = True
    else:
        partition_args.keygen = False
    partition_args.key_protect_hmac = False
    generate(partition_args)


def generate_onboarding_data(args, index, discriminator, passcode):
    payloads = SetupPayload(discriminator, passcode, args.discovery_mode, CommissioningFlow(args.commissioning_flow),
                            args.vendor_id, args.product_id)
    chip_qrcode = payloads.generate_qrcode()
    chip_manualcode = payloads.generate_manualcode()
    # ToDo: remove this if qrcode tool can handle the standard manual code format
    if args.commissioning_flow == CommissioningFlow.Standard:
        chip_manualcode = chip_manualcode[:4] + '-' + chip_manualcode[4:7] + '-' + chip_manualcode[7:]
    else:
        chip_manualcode = '"' + chip_manualcode[:4] + '-' + chip_manualcode[4:7] + '-' + chip_manualcode[7:11] + '\n' + chip_manualcode[11:15] + '-' + chip_manualcode[15:18] + '-' + chip_manualcode[18:20] + '-' + chip_manualcode[20:21] + '"'

    logging.info('Generated QR code: ' + chip_qrcode)
    logging.info('Generated manual code: ' + chip_manualcode)

    csv_data = 'qrcode,manualcode,discriminator,passcode\n'
    csv_data += chip_qrcode + ',' + chip_manualcode + ',' + str(discriminator) + ',' + str(passcode) + '\n'

    onboarding_data_file = os.sep.join([OUT_DIR['top'], UUIDs[index], '{}-onb_codes.csv'.format(UUIDs[index])])
    with open(onboarding_data_file, 'w') as f:
        f.write(csv_data)

    # Create QR code image as mentioned in the spec
    qrcode_file = os.sep.join([OUT_DIR['top'], UUIDs[index], '{}-qrcode.png'.format(UUIDs[index])])
    chip_qr = pyqrcode.create(chip_qrcode, version=2, error='M')
    chip_qr.png(qrcode_file, scale=6)

    logging.info('Generated onboarding data and QR Code')


def get_args():
    def any_base_int(s): return int(s, 0)

    parser = argparse.ArgumentParser(description='Manufacuring partition generator tool',
                                     formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=50))

    g_gen = parser.add_argument_group('General options')
    g_gen.add_argument('-n', '--count', type=any_base_int, default=1,
                       help='The number of manufacturing partition binaries to generate. Default is 1. \
                              If --csv and --mcsv are present, the number of lines in the mcsv file is used.')

    g_gen.add_argument('--target', default='esp32',
                       help='The platform type of device. eg: one of esp32, esp32c3, etc.')
    g_gen.add_argument('-s', '--size', type=any_base_int, default=0x6000,
                       help='The size of manufacturing partition binaries to generate. Default is 0x6000.')
    g_gen.add_argument('-e', '--encrypt', action='store_true', required=False,
                      help='Encrypt the factory parititon NVS binary')
    g_gen.add_argument('--log-level', default='info', choices=__LOG_LEVELS__.keys(),
                      help='Set the log level (default: %(default)s)')
    g_gen.add_argument('--outdir', default=os.path.join(os.getcwd(), 'out'),
                      help='The output directory for the generated files (default: %(default)s)')
    g_gen.add_argument('--no-bin', action='store_false', dest='generate_bin',
                        help='Do not generate the factory partition binary')

    g_commissioning = parser.add_argument_group('Commisioning options')
    g_commissioning.add_argument('--passcode', type=any_base_int,
                                 help='The passcode for pairing. Randomly generated if not specified.')
    g_commissioning.add_argument('--discriminator', type=any_base_int,
                                 help='The discriminator for pairing. Randomly generated if not specified.')
    g_commissioning.add_argument('-cf', '--commissioning-flow', type=any_base_int, default=0,
                                 help='Device commissioning flow, 0:Standard, 1:User-Intent, 2:Custom. \
                                          Default is 0.', choices=[0, 1, 2])
    g_commissioning.add_argument('-dm', '--discovery-mode', type=any_base_int, default=2,
                                 help='The discovery mode for commissionable device discovery. \
                                        2: BLE, 4: On-network, 6: BLE + On-network. Default is BLE.')
    g_commissioning.add_argument('--enable-dynamic-passcode', action="store_true", required=False,
                                 help='Enable dynamic passcode. If enabling this option, the generated binaries will \
                                         not include the spake2p verifier. so this option should work with a custom \
                                         CommissionableDataProvider which can generate random passcode and \
                                         corresponding verifier')

    g_dac = parser.add_argument_group('Device attestation credential options')
    g_dac.add_argument('--dac-in-secure-cert', action="store_true", required=False,
                        help='Store DAC in secure cert partition. By default, DAC is stored in nvs factory partition.')
    g_dac.add_argument('-lt', '--lifetime', default=36500, type=any_base_int,
                       help='Lifetime of the generated certificate. Default is 100 years if not specified, \
                              this indicate that certificate does not have well defined expiration date.')
    g_dac.add_argument('-vf', '--valid-from',
                       help='The start date for the certificate validity period in format <YYYY>-<MM>-<DD> [ <HH>:<MM>:<SS> ]. \
                              Default is current date.')
    g_dac.add_argument('-cn', '--cn-prefix', default='ESP32',
                       help='The common name prefix of the subject of the PAI certificate.')
    # If DAC is present then PAI key is not required, so it is marked as not required here
    # but, if DAC is not present then PAI key is required and that case is validated in validate_args()
    g_dac.add_argument('-c', '--cert', help='The input certificate file in PEM format.')
    g_dac.add_argument('-k', '--key', help='The input key file in PEM format.')
    g_dac.add_argument('-cd', '--cert-dclrn', help='The certificate declaration file in DER format.')
    g_dac.add_argument('--dac-cert', help='The input DAC certificate file in PEM format.')
    g_dac.add_argument('--dac-key', help='The input DAC private key file in PEM format.')
    g_dac.add_argument('-ds', '--ds-peripheral', action="store_true",
                       help='Use DS Peripheral in generating secure cert partition.')
    g_dac.add_argument('--efuse-key-id', type=int, choices=range(0, 6), default=-1,
                        help='Provide the efuse key_id which contains/will contain HMAC_KEY, default is 1')
    g_dac.add_argument('--port', dest='port', help='UART com port to which the ESP device is connected')
    g_dac.add_argument('--pwd', '--password', dest='priv_key_pass', help='The password associated with the private key')

    input_cert_group = g_dac.add_mutually_exclusive_group(required=False)
    input_cert_group.add_argument('--paa', action='store_true', help='Use input certificate as PAA certificate.')
    input_cert_group.add_argument('--pai', action='store_true', help='Use input certificate as PAI certificate.')

    g_dev_inst_info = parser.add_argument_group('Device instance information options')
    g_dev_inst_info.add_argument('-v', '--vendor-id', type=any_base_int, help='Vendor id', required=True)
    g_dev_inst_info.add_argument('--vendor-name', help='Vendor name')
    g_dev_inst_info.add_argument('-p', '--product-id', type=any_base_int, help='Product id', required=True)
    g_dev_inst_info.add_argument('--product-name', help='Product name')
    g_dev_inst_info.add_argument('--hw-ver', type=any_base_int, help='Hardware version')
    g_dev_inst_info.add_argument('--hw-ver-str', help='Hardware version string')
    g_dev_inst_info.add_argument('--mfg-date', help='Manufacturing date in format YYYY-MM-DD')
    g_dev_inst_info.add_argument('--serial-num', help='Serial number')
    g_dev_inst_info.add_argument('--enable-rotating-device-id', action='store_true', help='Enable Rotating device id in the generated binaries')
    g_dev_inst_info.add_argument('--rd-id-uid',
                        help='128-bit unique identifier for generating rotating device identifier, provide 32-byte hex string, e.g. "1234567890abcdef1234567890abcdef"')
    product_finish_choices = [finish.name for finish in ProductFinish]
    g_dev_inst_info.add_argument("--product-finish", type=str, choices=product_finish_choices,
                        help='Product finishes choices for product appearance')

    product_color_choices = [color.name for color in ProductColor]
    g_dev_inst_info.add_argument("--product-color", type=str, choices=product_color_choices,
                        help='Product colors choices for product appearance')

    g_dev_inst_info.add_argument("--part-number", type=str, help='human readable product number')


    g_dev_inst = parser.add_argument_group('Device instance options')
    g_dev_inst.add_argument('--calendar-types', nargs='+',
                            help='List of supported calendar types. Supported Calendar Types: Buddhist, Chinese, Coptic, \
                                Ethiopian, Gregorian, Hebrew, Indian, Islamic, Japanese, Korean, Persian, Taiwanese')
    g_dev_inst.add_argument('--locales', nargs='+',
                            help='List of supported locales, Language Tag as defined by BCP47, eg. en-US en-GB')
    g_dev_inst.add_argument('--fixed-labels', nargs='+',
                            help='List of fixed labels, eg: "0/orientation/up" "1/orientation/down" "2/orientation/down"')

    g_dev_inst.add_argument('--supported-modes', type=str, nargs='+', required=False,
                        help='List of supported modes, eg: mode1/label1/ep/"tagValue1\\mfgCode, tagValue2\\mfgCode"  mode2/label2/ep/"tagValue1\\mfgCode, tagValue2\\mfgCode"  mode3/label3/ep/"tagValue1\\mfgCode, tagValue2\\mfgCode"')

    g_basic = parser.add_argument_group('Few more Basic clusters options')
    g_basic.add_argument('--product-label', help='Product label')
    g_basic.add_argument('--product-url', help='Product URL')

    g_extra_info = parser.add_argument_group('Extra information options using csv files')
    g_extra_info.add_argument('--csv', help='CSV file containing the partition schema for extra options. \
            [REF: https://docs.espressif.com/projects/esp-idf/en/latest/esp32/api-reference/storage/mass_mfg.html#csv-configuration-file]')
    g_extra_info.add_argument('--mcsv', help='Master CSV file containig optional/extra values specified by the user. \
            [REF: https://docs.espressif.com/projects/esp-idf/en/latest/esp32/api-reference/storage/mass_mfg.html#master-value-csv-file]')

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    return parser.parse_args()


def add_optional_KVs(args):
    # Device instance information
    if args.vendor_id is not None:
        chip_factory_append('vendor-id', 'data', 'u32', args.vendor_id)
    if args.vendor_name is not None:
        chip_factory_append('vendor-name', 'data', 'string', args.vendor_name)
    if args.product_id is not None:
        chip_factory_append('product-id', 'data', 'u32', args.product_id)
    if args.product_name is not None:
        chip_factory_append('product-name', 'data', 'string', args.product_name)
    if args.hw_ver is not None:
        chip_factory_append('hardware-ver', 'data', 'u32', args.hw_ver)
    if args.hw_ver_str is not None:
        chip_factory_append('hw-ver-str', 'data', 'string', args.hw_ver_str)
    if args.mfg_date is not None:
        chip_factory_append('mfg-date', 'data', 'string', args.mfg_date)
    if args.enable_rotating_device_id:
        chip_factory_append('rd-id-uid', 'data', 'hex2bin', args.rd_id_uid)
    if args.product_finish:
        chip_factory_append('product-finish', 'data', 'u32', ProductFinish[args.product_finish].value)
    if args.product_color:
        chip_factory_append('product-color', 'data', 'u32', ProductColor[args.product_color].value)
    if args.part_number:
        chip_factory_append('part-number', 'data', 'string', args.part_number)

    # Add the serial-num
    chip_factory_append('serial-num', 'data', 'string', args.serial_num)

    # Add certificates and keys
    if (args.paa or args.pai) and (not args.dac_in_secure_cert):
        chip_factory_append('dac-cert', 'file', 'binary', None)
        chip_factory_append('dac-key', 'file', 'binary', None)
        chip_factory_append('dac-pub-key', 'file', 'binary', None)
        chip_factory_append('pai-cert', 'file', 'binary', None)

    if not args.enable_dynamic_passcode:
        chip_factory_append('verifier', 'data', 'string', None)

    # Add certificate declaration
    if args.cert_dclrn:
        chip_factory_append('cert-dclrn','file','binary', os.path.relpath(args.cert_dclrn))

    # Add the Keys in csv files
    if args.csv is not None:
        chip_nvs_map_append_config_csv(args.csv)

    # Device information
    if args.calendar_types is not None:
        chip_factory_append('cal-types', 'data', 'u32', calendar_types_to_uint32(args.calendar_types))

    # Supported locale is stored as multiple entries, key format: "locale/<index>, example key: "locale/0"
    if (args.locales is not None):
        chip_factory_append('locale-sz', 'data', 'u32', len(args.locales))
        for i in range(len(args.locales)):
            chip_factory_append('locale/{:x}'.format(i), 'data', 'string', args.locales[i])

    # Each endpoint can contains the fixed lables
    #  - fl-sz/<index>     : number of fixed labels for the endpoint
    #  - fl-k/<ep>/<index> : fixed label key for the endpoint and index
    #  - fl-v/<ep>/<index> : fixed label value for the endpoint and index
    if (args.fixed_labels is not None):
        dict = get_fixed_label_dict(args.fixed_labels)
        for key in dict.keys():
            chip_factory_append('fl-sz/{:x}'.format(int(key)), 'data', 'u32', len(dict[key]))

            for i in range(len(dict[key])):
                entry = dict[key][i]
                chip_factory_append('fl-k/{:x}/{:x}'.format(int(key), i), 'data', 'string', list(entry.keys())[0])
                chip_factory_append('fl-v/{:x}/{:x}'.format(int(key), i), 'data', 'string', list(entry.values())[0])

    # SupportedModes are stored as multiple entries
    #  - sm-sz/<ep>                 : number of supported modes for the endpoint
    #  - sm-label/<ep>/<index>      : supported modes label key for the endpoint and index
    #  - sm-mode/<ep>/<index>       : supported modes mode key for the endpoint and index
    #  - sm-st-sz/<ep>/<index>      : supported modes SemanticTag key for the endpoint and index
    #  - st-v/<ep>/<index>/<ind>    : semantic tag value key for the endpoint and index and ind
    #  - st-mfg/<ep>/<index>/<ind>  : semantic tag mfg code key for the endpoint and index and ind
    if (args.supported_modes is not None):
        dictionary = get_supported_modes_dict(args.supported_modes)
        for ep in dictionary.keys():
            chip_factory_append('sm-sz/{:x}'.format(int(ep)), 'data', 'u32', len(dictionary[ep]))

            for i in range(len(dictionary[ep])):
                item = dictionary[ep][i]

                chip_factory_append('sm-label/{:x}/{:x}'.format(int(ep), i), 'data', 'string', item["Label"])
                chip_factory_append('sm-mode/{:x}/{:x}'.format(int(ep), i), 'data', 'u32', item["Mode"])
                chip_factory_append('sm-st-sz/{:x}/{:x}'.format(int(ep), i), 'data', 'u32', len(item["Semantic_Tag"]))

                for j in range(len(item["Semantic_Tag"])):
                    entry = item["Semantic_Tag"][j]

                    _value = {
                        'type': 'data',
                        'encoding': 'u32',
                        'value': entry["value"]
                    }
                    _mfg_code = {
                        'type': 'data',
                        'encoding': 'u32',
                        'value': entry["mfgCode"]
                    }

                    chip_factory_append('st-v/{:x}/{:x}/{:x}'.format(int(ep), i, j), 'data', 'u32', entry["value"])
                    chip_factory_append('st-mfg/{:x}/{:x}/{:x}'.format(int(ep), i, j), 'data', 'u32', entry["mfgCode"])

    # Keys from basic clusters
    if args.product_label is not None:
        chip_factory_append('product-label', 'data', 'string', args.product_label)
    if args.product_url is not None:
        chip_factory_append('product-url', 'data', 'string', args.product_url)

def main_internal(args):
    logging.basicConfig(format='[%(asctime)s] [%(levelname)7s] - %(message)s', level=__LOG_LEVELS__[args.log_level])
    validate_args(args)
    setup_out_dirs(args.vendor_id, args.product_id, args.count, args.outdir, args.dac_cert)
    add_optional_KVs(args)
    generate_passcodes_and_discriminators(args)
    write_csv_files(args)
    if args.paa or args.pai:
        setup_root_certs(args)
    write_per_device_unique_data(args)
    generate_partitions('matter_partition', args.size, args.encrypt, args.generate_bin)
    organize_output_files('matter_partition', args)
    generate_summary(args)

def main():
    args = get_args()
    main_internal(args)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

# Copyright 2022 Espressif Systems (Shanghai) PTE LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains the CHIP specific key along with the data type and encoding format
"""

# This map contains mandatory CHIP specific key along with the data type and encoding format
# Additionally to add more keys to chip-factory, use chip_factory_append() API

import csv

CHIP_NVS_MAP = {
    'chip-factory': {
        # Commissionable Data
        'discriminator': {
            'type': 'data',
            'encoding': 'u32',
            'value': None,
        },
        'iteration-count': {
            'type': 'data',
            'encoding': 'u32',
            'value': None,
        },
        'salt': {
            'type': 'data',
            'encoding': 'string',
            'value': None,
        },
    }
}


def get_dict(key, type, encoding, value):
    return {
        key: {
            'type': type,
            'encoding': encoding,
            'value': value,
        }
    }


def get_namespace_dict(namespace):
    return {
        namespace: {
        }
    }


def chip_nvs_get_config_csv():
    csv_data = ''
    for k, v in CHIP_NVS_MAP.items():
        csv_data += k + ',' + 'namespace,' + '\n'
        for k1, v1 in v.items():
            csv_data += k1 + ',' + v1['type'] + ',' + v1['encoding'] + '\n'
    return csv_data


def chip_factory_append(key, type, encoding, value):
    CHIP_NVS_MAP['chip-factory'].update(get_dict(key, type, encoding, value))


def chip_factory_update(key, value):
    CHIP_NVS_MAP['chip-factory'][key]['value'] = value


def chip_nvs_map_update(namespace, key, type, encoding, value):
    CHIP_NVS_MAP[namespace].update(get_dict(key, type, encoding, value))

def chip_nvs_map_append_config_csv(csv_path):
    with open(csv_path, 'r') as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        # Set current namespace to 'chip-factory' when the first line of the csv file is not 'XX,namespace,'
        current_namespace = 'chip-factory'
        for csv_data in csv_reader:
            if 'namespace' in csv_data:
                current_namespace = csv_data[0]
                if (current_namespace not in list(CHIP_NVS_MAP.keys())):
                    CHIP_NVS_MAP.update(get_namespace_dict(current_namespace))
            else:
                chip_nvs_map_update(current_namespace, csv_data[0], csv_data[1], csv_data[2], None)


def chip_factory_get_val(key):
    return CHIP_NVS_MAP['chip-factory'][key]['value']


def chip_get_keys_as_csv():
    keys = list()
    for ns in CHIP_NVS_MAP:
        keys.extend(list(CHIP_NVS_MAP[ns]))
    return ','.join(keys)


def chip_get_values_as_csv():
    values = list()
    for ns in CHIP_NVS_MAP:
        for k in CHIP_NVS_MAP[ns]:
            values.append(str(CHIP_NVS_MAP[ns][k]['value']))
    return ','.join(values)



#!/usr/bin/env python3

# Copyright 2022 Espressif Systems (Shanghai) PTE LTD
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains utilitiy functions for validating argument.
"""

import sys
import enum
import logging
import subprocess
from bitarray import bitarray
from bitarray.util import ba2int

ROTATING_DEVICE_ID_UNIQUE_ID_LEN_BITS = 128
SERIAL_NUMBER_LEN = 16


# Lengths for manual pairing codes and qrcode
SHORT_MANUALCODE_LEN = 11
LONG_MANUALCODE_LEN = 21
QRCODE_LEN = 22


INVALID_PASSCODES = [00000000, 11111111, 22222222, 33333333, 44444444, 55555555,
                     66666666, 77777777, 88888888, 99999999, 12345678, 87654321]


class CalendarTypes(enum.Enum):
    Buddhist = 0
    Chinese = 1
    Coptic = 2
    Ethiopian = 3
    Gregorian = 4
    Hebrew = 5
    Indian = 6
    Islamic = 7
    Japanese = 8
    Korean = 9
    Persian = 10
    Taiwanese = 11

class ProductFinish(enum.Enum):
    other = 0
    matte = 1
    satin = 2
    polished = 3
    rugged = 4
    fabric = 5


class ProductColor(enum.Enum):
    black = 0
    navy = 1
    green = 2
    teal = 3
    maroon = 4
    purple = 5
    olive = 6
    gray = 7
    blue = 8
    lime = 9
    aqua = 10
    red = 11
    fuchsia = 12
    yellow = 13
    white = 14
    nickel = 15
    chrome = 16
    brass = 17
    copper = 18
    silver = 19
    gold = 20


def vid_pid_str(vid, pid):
    return '_'.join([hex(vid)[2:], hex(pid)[2:]])


def disc_pin_str(discriminator, passcode):
    return '_'.join([hex(discriminator)[2:], hex(passcode)[2:]])


# Checks if the input string is a valid hex string
def ishex(s):
    try:
        n = int(s, 16)
        return True
    except ValueError:
        return False


# Validate the input string length against the min and max length
def check_str_range(s, min_len, max_len, name):
    if s and ((len(s) < min_len) or (len(s) > max_len)):
        logging.error('%s must be between %d and %d characters', name, min_len, max_len)
        sys.exit(1)


# Validate the input integer range
def check_int_range(value, min_value, max_value, name):
    if value and ((value < min_value) or (value > max_value)):
        logging.error('%s is out of range, should be in range [%d, %d]', name, min_value, max_value)
        sys.exit(1)


# Validates discriminator and passcode
def validate_commissionable_data(args):
    check_int_range(args.discriminator, 0x0000, 0x0FFF, 'Discriminator')
    check_int_range(args.discovery_mode, 0, 7, 'Discovery mode')
    if args.passcode is not None:
        if ((args.passcode < 0x0000001 and args.passcode > 0x5F5E0FE) or (args.passcode in INVALID_PASSCODES)):
            logging.error('Invalid passcode' + str(args.passcode))
            sys.exit(1)


# Validate the device instance information
def validate_device_instance_info(args):
    check_int_range(args.product_id, 0x0000, 0xFFFF, 'Product id')
    check_int_range(args.vendor_id, 0x0000, 0xFFFF, 'Vendor id')
    check_int_range(args.hw_ver, 0x0000, 0xFFFF, 'Hardware version')
    check_str_range(args.serial_num, 1, SERIAL_NUMBER_LEN, 'Serial number')
    check_str_range(args.vendor_name, 1, 32, 'Vendor name')
    check_str_range(args.product_name, 1, 32, 'Product name')
    check_str_range(args.hw_ver_str, 1, 64, 'Hardware version string')
    check_str_range(args.mfg_date, 8, 16, 'Manufacturing date')
    check_str_range(args.rd_id_uid, 32, 32, 'Rotating device Unique id')


# Validate the device information: calendar types and fixed labels
def validate_device_info(args):
    # Validate the input calendar types
    if args.calendar_types is not None:
        if not (set(args.calendar_types) <= set(CalendarTypes.__members__)):
            invalid_types = set(args.calendar_types).union(set(CalendarTypes.__members__)) - set(CalendarTypes.__members__)
            logging.error('Unknown calendar type/s: %s', invalid_types)
            logging.error('Supported calendar types: %s', ', '.join(CalendarTypes.__members__))
            sys.exit(1)

    if args.fixed_labels is not None:
        for fl in args.fixed_labels:
            _l = fl.split('/')
            if len(_l) != 3:
                logging.error('Invalid fixed label: %s', fl)
                sys.exit(1)

            if not (ishex(_l[0]) and (len(_l[1]) > 0 and len(_l[1]) < 16) and (len(_l[2]) > 0 and len(_l[2]) < 16)):
                logging.error('Invalid fixed label: %s', fl)
                sys.exit(1)


# Validates the attestation related arguments
def validate_attestation_info(args):
    # DAC key and DAC cert both should be present or none
    if (args.dac_key is not None) != (args.dac_cert is not None):
        logging.error("dac_key and dac_cert should be both present or none")
        sys.exit(1)
    else:
        # Make sure PAI certificate is present if DAC is present
        if (args.dac_key is not None) and (args.pai is False):
            logging.error('Please provide PAI certificate along with DAC certificate and DAC key')
            sys.exit(1)

    # Validate the input certificate type, if DAC is not present
    if args.dac_key is None and args.dac_cert is None:
        if args.paa:
            logging.info('Input Root certificate type PAA')
        elif args.pai:
            logging.info('Input Root certificate type PAI')
        else:
            logging.info('Do not include the device attestation certificates and keys in partition binaries')

        # Check if Key and certificate are present
        if (args.paa or args.pai) and (args.key is None or args.cert is None):
            logging.error('CA key and certificate are required to generate DAC key and certificate')
            sys.exit(1)


# Validates few basic cluster related arguments: product-label and product-url
def validate_basic_cluster_info(args):
    check_str_range(args.product_label, 1, 64, 'Product Label')
    check_str_range(args.product_url, 1, 256, 'Product URL')


# Validates the input arguments, this calls the above functions
def validate_args(args):
    # csv and mcsv both should present or none
    if (args.csv is not None) != (args.mcsv is not None):
        logging.error("csv and mcsv should be both present or none")
        sys.exit(1)
    else:
        # Read the number of lines in mcsv file
        if args.mcsv is not None:
            with open(args.mcsv, 'r') as f:
                lines = sum(1 for line in f)

            # Subtract 1 for the header line
            args.count = lines - 1

    validate_commissionable_data(args)
    validate_device_instance_info(args)
    validate_device_info(args)
    validate_attestation_info(args)
    validate_basic_cluster_info(args)

    # If discriminator/passcode/DAC/serial_number/rotating_device_id is present
    # then we are restricting the number of partitions to 1
    if (args.discriminator is not None
            or args.passcode is not None
            or args.dac_key is not None
            or args.serial_num is not None
            or args.rd_id_uid is not None):
        if args.count > 1:
            logging.error('Number of partitions should be 1 when discriminator or passcode or DAC or serial number or rotating device id is present')
            sys.exit(1)

    logging.info('Number of manufacturing NVS images to generate: {}'.format(args.count))


# Supported Calendar types is stored as a bit array in one uint32_t.
def calendar_types_to_uint32(calendar_types):
    # In validate_device_info() we have already verified that the calendar types are valid
    result = bitarray(32, endian='little')
    result.setall(0)
    for calendar_type in calendar_types:
        result[CalendarTypes[calendar_type].value] = 1
    return ba2int(result)


# get_fixed_label_dict() converts the list of strings to per endpoint dictionaries.
# example input  : ['0/orientation/up', '1/orientation/down', '2/orientation/down']
# example outout : {'0': [{'orientation': 'up'}], '1': [{'orientation': 'down'}], '2': [{'orientation': 'down'}]}
def get_fixed_label_dict(fixed_labels):
    fl_dict = {}
    for fl in fixed_labels:
        _l = fl.split('/')

        if len(_l) != 3:
            logging.error('Invalid fixed label: %s', fl)
            sys.exit(1)

        if not (ishex(_l[0]) and (len(_l[1]) > 0 and len(_l[1]) < 16) and (len(_l[2]) > 0 and len(_l[2]) < 16)):
            logging.error('Invalid fixed label: %s', fl)
            sys.exit(1)

        if _l[0] not in fl_dict.keys():
            fl_dict[_l[0]] = list()

        fl_dict[_l[0]].append({_l[1]: _l[2]})

    return fl_dict


# get_supported_modes_dict() converts the list of strings to per endpoint dictionaries.
# example with semantic tags
# input  : ['0/label1/1/"1\0x8000, 2\0x8000" 1/label2/1/"1\0x8000, 2\0x8000"']
# outout : {'1': [{'Label': 'label1', 'Mode': 0, 'Semantic_Tag': [{'value': 1, 'mfgCode': 32768}, {'value': 2, 'mfgCode': 32768}]}, {'Label': 'label2', 'Mode': 1, 'Semantic_Tag': [{'value': 1, 'mfgCode': 32768}, {'value': 2, 'mfgCode': 32768}]}]}

# example without semantic tags
# input  : ['0/label1/1 1/label2/1']
# outout : {'1': [{'Label': 'label1', 'Mode': 0, 'Semantic_Tag': []}, {'Label': 'label2', 'Mode': 1, 'Semantic_Tag': []}]}

def get_supported_modes_dict(supported_modes):
    output_dict = {}

    for mode_str in supported_modes:
        mode_label_strs = mode_str.split('/')
        mode = mode_label_strs[0]
        label = mode_label_strs[1]
        ep = mode_label_strs[2]

        semantic_tags = ''
        if (len(mode_label_strs) == 4):
            semantic_tag_strs = mode_label_strs[3].split(', ')
            semantic_tags = [{"value": int(v.split('\\')[0]), "mfgCode": int(v.split('\\')[1], 16)} for v in semantic_tag_strs]

        mode_dict = {"Label": label, "Mode": int(mode), "Semantic_Tag": semantic_tags}

        if ep in output_dict:
            output_dict[ep].append(mode_dict)
        else:
            output_dict[ep] = [mode_dict]

    return output_dict



#!/usr/bin/env python
#
# SPDX-FileCopyrightText: 2018-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
#

import argparse
import csv
import distutils.dir_util
import os
from itertools import zip_longest

import esp_idf_nvs_partition_gen.nvs_partition_gen as nvs_partition_gen


def create_temp_files(args):
    new_filenames = []
    for filename in [args.conf, args.values]:
        name, ext = os.path.splitext(filename)
        new_filename = name + '_tmp' + ext
        strip_blank_lines(filename, new_filename)
        new_filenames.append(new_filename)
    return new_filenames


def strip_blank_lines(input_filename, output_filename):
    with open(input_filename, 'r') as read_from, open(output_filename,'w', newline='') as write_to:
        for line in read_from:
            if not line.isspace():
                write_to.write(line)


def verify_values_exist(input_values_file, keys_in_values_file):
    """ Verify all keys have corresponding values in values file
    """
    with open(input_values_file, 'r') as values_file:
        values_file_reader = csv.reader(values_file, delimiter=',')
        next(values_file_reader)

        for line_num, line in enumerate(values_file_reader, start=2):
            key_count_in_values_file = len(keys_in_values_file)
            if len(line) != key_count_in_values_file:
                raise SystemExit('\nError: Number of values is not equal to number of keys in file: %s at line No:%s\n'
                                 % (str(input_values_file), str(line_num)))


def verify_keys_exist(values_file_keys, input_config_file):
    """ Verify all keys from config file are present in values file
    """
    keys_missing = []

    with open(input_config_file,'r') as config_file:
        config_file_reader = csv.reader(config_file, delimiter=',')

        for line_num, line in enumerate(config_file_reader, start=1):
            if 'namespace' not in line:
                if values_file_keys:
                    if line[0] == values_file_keys[0]:
                        del values_file_keys[0]
                    else:
                        keys_missing.append([line_num, line[0]])
                else:
                    keys_missing.append([line_num, line[0]])

        if keys_missing:
            for line_num, key in keys_missing:
                print('Key:`', str(key), '` at line no:', str(line_num),
                      ' in config file is not found in values file.')
            raise SystemExit(1)


def verify_datatype_encoding(input_config_file):
    """ Verify datatype and encodings from config file is valid
    """
    valid_encodings = {'string', 'binary', 'hex2bin','u8', 'i8', 'u16', 'u32', 'i32','base64'}
    valid_datatypes = {'file','data','namespace'}

    with open(input_config_file,'r') as config_file:
        config_file_reader = csv.reader(config_file, delimiter=',')

        for line_num, line in enumerate(config_file_reader, start=1):
            if line[1] not in valid_datatypes:
                raise SystemExit('Error: config file: %s has invalid datatype at line no:%s\n`'
                                 % (str(input_config_file), str(line_num)))
            if 'namespace' not in line and line[2] not in valid_encodings:
                raise SystemExit('Error: config file: %s has invalid encoding at line no:%s\n`'
                                 % (str(input_config_file), str(line_num)))


def verify_file_data_count(input_config_file, keys_repeat):
    """ Verify count of data on each line in config file is equal to 3
    (as format must be: <key,type and encoding>)
    """
    with open(input_config_file, 'r') as config_file:
        config_file_reader = csv.reader(config_file, delimiter=',')

        for line_num, line in enumerate(config_file_reader, start=1):
            if len(line) != 3 and line[0] not in keys_repeat:
                raise SystemExit('Error: data missing in config file at line no:%s <format needed:key,type,encoding>\n'
                                 % str(line_num))


def verify_data_in_file(input_config_file, input_values_file, config_file_keys, keys_in_values_file, keys_repeat):
    """ Verify count of data on each line in config file is equal to 3 \
    (as format must be: <key,type and encoding>)
    Verify datatype and encodings from config file is valid
    Verify all keys from config file are present in values file and \
    Verify each key has corresponding value in values file
    """
    try:
        verify_file_data_count(input_config_file, keys_repeat)

        verify_datatype_encoding(input_config_file)

        # Get keys from values file present in config files
        values_file_keys = get_keys(keys_in_values_file, config_file_keys)

        verify_keys_exist(values_file_keys, input_config_file)

        verify_values_exist(input_values_file, keys_in_values_file)

    except Exception as err:
        print(err)
        exit(1)


def get_keys(keys_in_values_file, config_file_keys):
    """ Get keys from values file present in config file
    """
    values_file_keys = [key for key in keys_in_values_file if key in config_file_keys]
    return values_file_keys


def add_config_data_per_namespace(input_config_file):
    """ Add config data per namespace to `config_data_to_write` list
    """
    config_data_to_write = []
    config_data_per_namespace = []

    with open(input_config_file,'r') as csv_config_file:
        config_file_reader = csv.reader(csv_config_file, delimiter=',')

        # `config_data_per_namespace` is added to `config_data_to_write` list after reading next namespace
        for config_data in config_file_reader:
            if 'REPEAT' in config_data:
                config_data.remove('REPEAT')
            if 'namespace' in config_data:
                if config_data_per_namespace:
                    config_data_to_write.append(config_data_per_namespace)
                    config_data_per_namespace = []
                    config_data_per_namespace.append(config_data)
                else:
                    config_data_per_namespace.append(config_data)
            else:
                config_data_per_namespace.append(config_data)

        # `config_data_per_namespace` is added to `config_data_to_write` list as EOF is reached
        if (not config_data_to_write) or (config_data_to_write and config_data_per_namespace):
            config_data_to_write.append(config_data_per_namespace)

    return config_data_to_write


def get_fileid_val(file_identifier, key_value_data, fileid_value):
    """ Get file identifier value
    """
    file_id_found = False

    for key in key_value_data:
        if file_identifier and not file_id_found and file_identifier in key:
            fileid_value = key[1]
            file_id_found = True

    if not file_id_found:
        fileid_value = str(int(fileid_value) + 1)

    return fileid_value


def add_data_to_file(config_data_to_write, key_value_pair, output_csv_file):
    """ Add data to csv target file
    """
    header = ['key', 'type', 'encoding', 'value']
    data_to_write = []

    with open(output_csv_file, 'w', newline='') as target_csv_file:
        output_file_writer = csv.writer(target_csv_file, delimiter=',')
        output_file_writer.writerow(header)

        for namespace_config_data in config_data_to_write:
            for data in namespace_config_data:
                data_to_write = data[:]
                if 'namespace' in data:
                    data_to_write.append('')
                    output_file_writer.writerow(data_to_write)
                else:
                    key = data[0]
                    while key not in key_value_pair[0]:
                        del key_value_pair[0]
                    if key in key_value_pair[0]:
                        value = key_value_pair[0][1]
                        data_to_write.append(value)
                        del key_value_pair[0]
                        output_file_writer.writerow(data_to_write)


def create_dir(filetype, output_dir_path):
    """ Create new directory(if doesn't exist) to store file generated
    """
    output_target_dir = os.path.join(output_dir_path,filetype,'')
    if not os.path.isdir(output_target_dir):
        distutils.dir_util.mkpath(output_target_dir)

    return output_target_dir


def set_repeat_value(total_keys_repeat, keys, csv_file, target_filename):
    with open(csv_file, 'r') as read_from, open(target_filename,'w', newline='') as write_to:
        csv_file_reader = csv.reader(read_from, delimiter=',')
        headers = next(csv_file_reader)
        values = next(csv_file_reader)
        csv_file_writer = csv.writer(write_to, delimiter=',')
        csv_file_writer.writerow(headers)
        csv_file_writer.writerow(values)

        # read new data, add value if key has repeat tag, write to new file
        for row in csv_file_reader:
            key_val_new = list(zip_longest(keys, row))
            key_val_pair = list(zip_longest(keys, values))
            key_repeated = total_keys_repeat[:]
            index = 0
            while key_val_new and key_repeated:
                #  if key has repeat tag, get its corresponding value, write to file
                if key_val_new[0][0] == key_repeated[0]:
                    row[index] = key_val_pair[0][1]
                    del key_repeated[0]
                del key_val_new[0]
                del key_val_pair[0]
                index += 1
            csv_file_writer.writerow(row)

    return target_filename


def create_intermediate_csv(args, keys_in_values_file, keys_repeat, is_encr=False):
    file_identifier_value = '0'
    # Add config data per namespace to `config_data_to_write` list
    config_data_to_write = add_config_data_per_namespace(args.conf)

    try:
        with open(args.values, 'r') as csv_values_file:
            values_file_reader = csv.reader(csv_values_file, delimiter=',')
            keys = next(values_file_reader)

        filename, file_ext = os.path.splitext(args.values)
        target_filename = filename + '_created' + file_ext
        if keys_repeat:
            target_values_file = set_repeat_value(keys_repeat, keys, args.values, target_filename)
        else:
            target_values_file = args.values

        with open(target_values_file, 'r') as csv_values_file:
            values_file_reader = csv.reader(csv_values_file, delimiter=',')
            next(values_file_reader)

            # Create new directory(if doesn't exist) to store csv file generated
            output_csv_target_dir = create_dir('csv', args.outdir)
            # Create new directory(if doesn't exist) to store bin file generated
            output_bin_target_dir = create_dir('bin', args.outdir)

            for values_data_line in values_file_reader:
                key_value_data = list(zip_longest(keys_in_values_file, values_data_line))

                # Get file identifier value from values file
                file_identifier_value = get_fileid_val(args.fileid, key_value_data, file_identifier_value)

                key_value_pair = key_value_data[:]

                # Verify if output csv file does not exist
                csv_filename = args.prefix + '-' + file_identifier_value + '.' + 'csv'
                output_csv_file = output_csv_target_dir + csv_filename
                if os.path.isfile(output_csv_file):
                    raise SystemExit('Target csv file: %s already exists.`' % output_csv_file)

                # Add values corresponding to each key to csv intermediate file
                add_data_to_file(config_data_to_write, key_value_pair, output_csv_file)
                print('\nCreated CSV file: ===>', output_csv_file)

                # Verify if output bin file does not exist
                bin_filename = args.prefix + '-' + file_identifier_value + '.' + 'bin'
                output_bin_file = output_bin_target_dir + bin_filename
                if os.path.isfile(output_bin_file):
                    raise SystemExit('Target binary file: %s already exists.`' % output_bin_file)

                args.input = output_csv_file
                args.output = os.path.join('bin', bin_filename)
                if args.keygen:
                    args.keyfile = 'keys-' + args.prefix + '-' + file_identifier_value

                if is_encr:
                    nvs_partition_gen.encrypt(args)
                else:
                    nvs_partition_gen.generate(args)

            print('\nFiles generated in %s ...' % args.outdir)

    except Exception as e:
        print(e)
        exit(1)


def verify_file_format(args):
    keys_in_config_file = []
    keys_in_values_file = []
    keys_repeat = []

    # Verify files have csv extension
    conf_name, conf_extension = os.path.splitext(args.conf)
    if conf_extension != '.csv':
        raise SystemExit('Error: config file: %s does not have the .csv extension.' % args.conf)
    values_name, values_extension = os.path.splitext(args.values)
    if values_extension != '.csv':
        raise SystemExit('Error: values file: %s does not have the .csv extension.' % args.values)

    # Verify files are not empty
    if os.stat(args.conf).st_size == 0:
        raise SystemExit('Error: config file: %s is empty.' % args.conf)
    if os.stat(args.values).st_size == 0:
        raise SystemExit('Error: values file: %s is empty.' % args.values)

    # Extract keys from config file
    with open(args.conf, 'r') as config_file:
        config_file_reader = csv.reader(config_file, delimiter=',')
        for config_data in config_file_reader:
            if 'namespace' not in config_data:
                keys_in_config_file.append(config_data[0])
            if 'REPEAT' in config_data:
                keys_repeat.append(config_data[0])

    # Extract keys from values file
    with open(args.values, 'r') as values_file:
        values_file_reader = csv.reader(values_file, delimiter=',')
        keys_in_values_file = next(values_file_reader)

    # Verify file identifier exists in values file
    if args.fileid:
        if args.fileid not in keys_in_values_file:
            raise SystemExit('Error: target_file_identifier: %s does not exist in values file.\n' % args.fileid)
    else:
        args.fileid = 1

    return keys_in_config_file, keys_in_values_file, keys_repeat


def generate(args):
    # Create work files with no blank lines
    args.conf, args.values = create_temp_files(args)

    # Verify input config and values file format
    keys_in_config_file, keys_in_values_file, keys_repeat = verify_file_format(args)

    # Verify data in the input_config_file and input_values_file
    verify_data_in_file(args.conf, args.values, keys_in_config_file,
                        keys_in_values_file, keys_repeat)

    encryption_enabled = False
    if (args.keygen or args.inputkey):
        encryption_enabled = True
        print('\nGenerating encrypted NVS binary images...')
    # Create intermediate csv file
    create_intermediate_csv(args, keys_in_values_file, keys_repeat, is_encr=encryption_enabled)


def generate_key(args):
    nvs_partition_gen.generate_key(args)


def main():
    try:
        parser = argparse.ArgumentParser(description='\nESP Manufacturing Utility', formatter_class=argparse.RawTextHelpFormatter)
        subparser = parser.add_subparsers(title='Commands',
                                          dest='command',
                                          help='\nRun mfg_gen.py {command} -h for additional help\n\n')

        parser_gen = subparser.add_parser('generate',
                                          help='Generate NVS partition',
                                          formatter_class=argparse.RawTextHelpFormatter)
        parser_gen.set_defaults(func=generate)
        parser_gen.add_argument('conf',
                                default=None,
                                help='Path to configuration csv file to parse')
        parser_gen.add_argument('values',
                                default=None,
                                help='Path to values csv file to parse')
        parser_gen.add_argument('prefix',
                                default=None,
                                help='Unique name for each output filename prefix')
        parser_gen.add_argument('size',
                                default=None,
                                help='Size of NVS partition in bytes\
                                    \n(must be multiple of 4096)')
        parser_gen.add_argument('--fileid',
                                default=None,
                                help='''Unique file identifier(any key in values file) \
                                    \nfor each filename suffix (Default: numeric value(1,2,3...)''')
        parser_gen.add_argument('--version',
                                choices=[1, 2],
                                default=2,
                                type=int,
                                help='''Set multipage blob version.\
                                    \nVersion 1 - Multipage blob support disabled.\
                                    \nVersion 2 - Multipage blob support enabled.\
                                    \nDefault: Version 2 ''')
        parser_gen.add_argument('--keygen',
                                action='store_true',
                                help='Generates key for encrypting NVS partition')
        parser_gen.add_argument('--keyfile',
                                default=None,
                                help=argparse.SUPPRESS)
        parser_gen.add_argument('--inputkey',
                                default=None,
                                help='File having key for encrypting NVS partition')
        parser_gen.add_argument('--outdir',
                                default=os.getcwd(),
                                help='Output directory to store files created\
                                    \n(Default: current directory)')
        parser_gen.add_argument('--key_protect_hmac',
                                action='store_true',
                                help='''If set, the NVS encryption key protection scheme based on HMAC\
                                \nperipheral is used; else the default scheme based on Flash Encryption\
                                \nis used''')
        parser_gen.add_argument('--kp_hmac_keygen',
                                action='store_true',
                                help='Generate the HMAC key for HMAC-based encryption scheme')
        parser_gen.add_argument('--kp_hmac_keyfile',
                                default=None,
                                help='Path to output HMAC key file')
        parser_gen.add_argument('--kp_hmac_inputkey',
                                default=None,
                                help='File having the HMAC key for generating the NVS encryption keys')
        parser_gen.add_argument('--input',
                                default=None,
                                help=argparse.SUPPRESS)
        parser_gen.add_argument('--output',
                                default=None,
                                help=argparse.SUPPRESS)
        parser_gen_key = subparser.add_parser('generate-key',
                                              help='Generate keys for encryption',
                                              formatter_class=argparse.RawTextHelpFormatter)
        parser_gen_key.set_defaults(func=generate_key)
        parser_gen_key.add_argument('--keyfile',
                                    default=None,
                                    help='Path to output encryption keys file')
        parser_gen_key.add_argument('--outdir',
                                    default=os.getcwd(),
                                    help='Output directory to store files created.\
                                        \n(Default: current directory)')
        parser_gen_key.add_argument('--key_protect_hmac',
                                    action='store_true',
                                    help='''If set, the NVS encryption key protection scheme based on HMAC\
                                    \nperipheral is used; else the default scheme based on Flash Encryption\
                                    \nis used''')
        parser_gen_key.add_argument('--kp_hmac_keygen',
                                    action='store_true',
                                    help='Generate the HMAC key for HMAC-based encryption scheme')
        parser_gen_key.add_argument('--kp_hmac_keyfile',
                                    default=None,
                                    help='Path to output HMAC key file')
        parser_gen_key.add_argument('--kp_hmac_inputkey',
                                    default=None,
                                    help='File having the HMAC key for generating the NVS encryption keys')

        args = parser.parse_args()
        args.func(args)

    except ValueError as err:
        print(err)
    except Exception as e:
        print(e)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#
#    Copyright (c) 2022 Project CHIP Authors
#    All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# TODO: Implement the decode method

CODES = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
         'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
         'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
         'U', 'V', 'W', 'X', 'Y', 'Z', '-', '.']
RADIX = len(CODES)
BASE38_CHARS_NEEDED_IN_CHUNK = [2, 4, 5]
MAX_BYTES_IN_CHUNK = 3


def encode(bytes):
    total_bytes = len(bytes)
    qrcode = ''

    for i in range(0, total_bytes, MAX_BYTES_IN_CHUNK):
        if (i + MAX_BYTES_IN_CHUNK) > total_bytes:
            bytes_in_chunk = total_bytes - i
        else:
            bytes_in_chunk = MAX_BYTES_IN_CHUNK

        value = 0
        for j in range(i, i + bytes_in_chunk):
            value = value + (bytes[j] << (8 * (j - i)))

        base38_chars_needed = BASE38_CHARS_NEEDED_IN_CHUNK[bytes_in_chunk - 1]
        while base38_chars_needed > 0:
            qrcode += CODES[int(value % RADIX)]
            value = int(value / RADIX)
            base38_chars_needed -= 1

    return qrcode



#-!/usr/bin/env python3
# SPDX-FileCopyrightText: 2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import os
import sys
import subprocess
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py
import pkg_resources

patch_list = [
{
    'package': 'deps',
    'file': 'patches/0001-mfg-gen-no-bin-and-replace-print-by-logging.patch'
}]

def is_patch_applied(patch_file):
    try:
        result = subprocess.run(["patch", "--dry-run", "-p2"], input=open(patch_file, "rb").read(),
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=False)
        return b'previously applied' in result.stderr or result.returncode != 0
    except Exception as e:
        print(f"Error checking patch status: {e}")
        return False

def apply_patch(pkg_name, file_name):
    patch_file = pkg_resources.resource_filename(pkg_name, file_name)

    if is_patch_applied(patch_file):
        print("Patch already applied, skipping.")
        return

    try:
        subprocess.run(["patch", "-p2"], input=open(patch_file, "rb").read(), check=True)
        print("Patch applied successfully.")
    except subprocess.CalledProcessError as e:
        print(f"Error applying patch: {e}")
        sys.exit(1)

class CustomBuild(build_py):
    def run(self):
        for patch in patch_list:
            print(f" Applying patch {patch['file']} before building the package ")
            # Apply patch before building
            apply_patch(patch['package'], patch['file'])
        # Continue with normal build process
        super().run()


def get_install_requires():
    with open(os.path.realpath('requirements.txt')) as f:
        required = f.read().splitlines()
        return required

try:
    from setuptools import find_packages, setup
except ImportError:
    print(
        "Package setuptools is missing from your Python installation. "
        "Please see the installation section in the esp-matter-mfg-tool "
        "documentation for instructions on how to install it."
    )
    exit(1)

VERSION = "1.0.9"

long_description = """
====================
esp-matter-mfg-tool
====================
The python utility helps to generate the matter manufacturing partitions.

Source code for `esp-matter-mfg-tool` is
`hosted on github <https://github.com/espressif/esp-matter-tools/tree/main/mfg_tool>`_.

Documentation
-------------
Visit online `esp-matter-mfg-tool documentation <https://github.com/espressif/esp-matter-tools/tree/main/mfg_tool>`_
or run ``esp-matter-mfg-tool -h``.

License
-------
The License for the project can be found
`here <https://github.com/espressif/esp-matter-tools/tree/main/LICENSE>`_
"""

setup(
    name = "esp-matter-mfg-tool",
    version = VERSION,
    description = "A python utility which helps to generate matter manufacturing partitions",
    long_description = long_description,
    long_description_content_type = 'text/x-rst',
    url = "https://github.com/espressif/esp-matter-tools/tree/main/mfg_tool",

    project_urls = {
        "Documentation": "https://github.com/espressif/esp-matter-tools/tree/main/mfg_tool/README.md",
        "Source": "https://github.com/espressif/esp-matter-tools/tree/main/mfg_tool",
    },

    author = "Espressif Systems",
    author_email = "",
    license = "Apache-2.0",

    classifiers = [
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Topic :: Software Development :: Embedded Systems",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],

    python_requires = ">=3.8",
    setup_requires = (["wheel"] if "bdist_wheel" in sys.argv else []),
    install_requires = get_install_requires(),
    include_package_data = True,
    packages = find_packages(),
    package_data = {
      'deps':['patches/*.patch'],
    },
    entry_points={
        'console_scripts': [
            'esp-matter-mfg-tool = sources.mfg_tool:main',
        ],
    },
    cmdclass={"build_py": CustomBuild},
)
